import inspect
import os
import sys
from types import GenericAlias

import strawberry
from fastapi import Request
from graphql.error import GraphQLError
from graphql.error.graphql_error import format_error as format_graphql_error
from strawberry.fastapi import GraphQLRouter
from strawberry.http import GraphQLHTTPResponse
from strawberry.types import ExecutionResult

from .dl import MyDataLoader
from .models import Graphemy
from .setup import Setup


async def fake_dl_one(keys):
    return {k: None for k in keys}.values()


async def fake_dl_list(keys):
    return {k: [] for k in keys}.values()


class Query:
    pass


class Mutation:
    pass


async def hello_world(self, info) -> str:
    return 'Hello World'


class GraphemyRouter(GraphQLRouter):
    functions = []

    def __init__(
        self,
        query: object = Query,
        mutation: object = Mutation,
        folder: str = '',
        context_getter: callable = None,
        permission_getter: callable = None,
        dl_filter: callable = None,
        query_filter: callable = None,
        engine=None,
        extensions: list = [],
        **kwargs,
    ):
        functions = []
        classes = {}
        classes_folder = {}
        Setup.setup(
            engine=engine,
            folder=folder,
            get_perm=permission_getter,
            query_filter=query_filter,
        )
        count = 0
        for root, dirs, files in os.walk(Setup.folder):
            for file in files:
                if file.endswith('.py') and file != '__init__.py':
                    count += 1
                    module_name = os.path.splitext(file)[0]
                    module_path = os.path.join(root, module_name)
                    module_path_rel = os.path.relpath(
                        os.path.join(root, module_name)
                    )
                    module_path = module_path_rel.replace(os.path.sep, '.')
                    exec(f'import {module_path}')
                    functions.extend(
                        [
                            (
                                n,
                                f,
                                module_path_rel,
                            )
                            for n, f in inspect.getmembers(
                                sys.modules[module_path], inspect.isfunction
                            )
                            if n.startswith('dl_')
                        ]
                    )
                    for n, cls in [
                        (n, cls)
                        for n, cls in inspect.getmembers(
                            sys.modules[module_path], inspect.isclass
                        )
                        if issubclass(cls, Graphemy) and n != 'Graphemy'
                    ]:
                        classes[n] = (cls, module_path_rel)
                        classes_folder[module_path_rel] = cls
        print(count, ' loaded files in ', os.getcwd())
        need_query = True
        need_mutation = True
        for n, (cls, path) in classes.items():
            cls.set_schema(classes)
            setattr(cls, 'folder', path)
            setattr(
                sys.modules[__name__],
                cls.__name__ + 'Schema',
                cls.schema,
            )
            setattr(
                sys.modules[__name__],
                cls.__name__ + 'Filter',
                cls.filter,
            )
            if not cls._disable_query.default:
                need_query = False
                setattr(
                    query,
                    cls.__query__
                    if hasattr(cls, '__query__')
                    else cls.__tablename__ + 's',
                    strawberry.field(
                        cls.query(), permission_classes=[cls.auth('query')]
                    ),
                )
            if cls._default_mutation.default:
                need_mutation = False
                temp = strawberry.mutation(
                    cls.mutation,
                    permission_classes=[cls.auth('mutation')],
                )
                setattr(
                    mutation,
                    'put_' + cls.__tablename__.lower(),
                    temp,
                )
            if cls._delete_mutation.default:
                setattr(
                    mutation,
                    'delete_' + cls.__tablename__.lower(),
                    strawberry.mutation(
                        cls.delete_mutation,
                        permission_classes=[cls.auth('delete_mutation')],
                    ),
                )
        if need_query:
            setattr(
                query,
                'hello_world',
                strawberry.field(hello_world),
            )

        async def get_context(request: Request) -> dict:
            context = await context_getter(request) if context_getter else {}
            for n, f, m in functions:
                cls = classes_folder[m] if m in classes_folder else None
                context[n] = MyDataLoader(
                    load_fn=f
                    if await Setup.get_permission(cls, context, 'query')
                    else fake_dl_list
                    if type(inspect.signature(f).return_annotation)
                    == GenericAlias
                    else fake_dl_one,
                    filter_method=dl_filter,
                    request=request,
                )
            return context

        schema = strawberry.Schema(
            query=strawberry.type(query),
            mutation=None
            if need_mutation and mutation == Mutation
            else strawberry.type(mutation),
            extensions=extensions,
        )
        super().__init__(schema=schema, context_getter=get_context, **kwargs)

    async def process_result(
        self, request: Request, result: ExecutionResult
    ) -> GraphQLHTTPResponse:
        data: GraphQLHTTPResponse = {'data': result.data}
        errors = []
        if 'errors' in request.scope:
            errors.append(
                format_graphql_error(
                    GraphQLError(
                        "User don't have necessary permissions for this path",
                        path=[
                            c.__tablename__ for c in request.scope['errors']
                        ],
                    )
                )
            )
        if result.errors:
            errors.extend([format_graphql_error(err) for err in result.errors])
        if len(errors) > 0:
            data['errors'] = errors
        return data



from strawberry.dataloader import DataLoader

from .models import MyDate


class MyDataLoader(DataLoader):
    def __init__(self, filter_method=None, request=None, **kwargs):
        self.filter_method = filter_method
        self.request = request
        super().__init__(**kwargs)

    async def load(self, keys, filters: dict | None = False):
        if filters == False:
            return await super().load(keys)
        filters['keys'] = (
            tuple(keys)
            if isinstance(keys, list)
            else keys.strip()
            if isinstance(keys, str)
            else keys
        )
        data = await super().load(dict_to_tuple(filters))
        if self.filter_method:
            data = self.filter_method(data, self.request)
        return data


def dict_to_tuple(data: dict) -> tuple:
    """
    Recursively converts a nested dictionary to a sorted tuple of key-value pairs.

    This function takes a dictionary as input and recursively traverses it, converting
    it into a sorted tuple of key-value pairs. If nested dictionaries or lists are
    encountered, they are also recursively converted into sorted tuples. The final
    result is a sorted tuple of all key-value pairs in the input dictionary.

    Args:
        data (dict): The input dictionary to be converted.

    Returns:
        tuple: A sorted tuple of key-value pairs from the input dictionary.

    Example:
        >>> data = {
        ...     'name': 'John',
        ...     'age': 30,
        ...     'address': {
        ...         'street': '123 Main St',
        ...         'city': 'Exampleville'
        ...     }
        ... }
        >>> dict_to_tuple(data)
        (('address', (('city', 'Exampleville'), ('street', '123 Main St'))), ('age', 30), ('name', 'John'))

    Note:
        This function is compatible with dictionaries that contain nested dictionaries,
        lists, and instances of 'MyDate' objects. Lists are converted to sorted tuples
        before further processing.

    See Also:
        - MyDate: An example custom data type that can be converted into a dictionary.

    """
    result = []
    for key, value in data.items():
        if isinstance(value, MyDate):
            value = vars(value)
        if isinstance(value, dict):
            nested_tuples = dict_to_tuple(value)
            result.append((key, nested_tuples))
        elif isinstance(value, list):
            # Substitutes the list with a tuple before recursively calling the function
            nested_tuples = tuple(
                sorted(
                    dict_to_tuple(item)
                    if isinstance(item, dict) or isinstance(item, MyDate)
                    else item
                    for item in value
                )
            )
            result.append((key, nested_tuples))
        else:
            result.append((key, value))
    return tuple(sorted(result))


def dl(class_name: str | None = None, many: bool = True) -> callable:
    """
    Decorator for attaching metadata to a function indicating the expected behavior of a DataLoader.

    Args:
        class_name (str, optional): The data type that the DataLoader is expected to return. Defaults to None.
        many (bool, optional): Indicates whether the DataLoader should return a single item or a list of items.
            Defaults to True.

    Returns:
        callable: The decorated function.

    Example:
        >>> @dl(class_name="MyData", many=False)
        ... def load_data(item_id: int) -> dict:
        ...     # Implementation of data loading
        ...     pass

        >>> assert getattr(load_data, 'dl') == "MyData"
        >>> assert getattr(load_data, 'many') is False
    """

    def wrapper(func):
        setattr(func, 'dl', class_name)
        setattr(func, 'many', many)
        return func

    return wrapper



from dotenv import load_dotenv
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import sessionmaker
from sqlmodel import Session
from strawberry.permission import BasePermission


class Setup:
    engine = None
    folder = ''
    get_permission = None
    async_engine = False

    @classmethod
    async def execute_query(cls, query):
        if cls.async_engine:
            async_session = sessionmaker(
                cls.engine, class_=AsyncSession, expire_on_commit=False
            )
            async with async_session() as session:
                r = await session.execute(query)
                return r.scalars().all()
        else:
            with Session(cls.engine) as session:
                r = session.exec(query).all()
                return r

    @classmethod
    def setup(cls, engine, folder=None, get_perm=None, query_filter=None):
        if engine and 'async' in engine.__module__:
            cls.async_engine = True
        cls.engine = engine
        if query_filter:
            cls.query_filter = query_filter
        else:

            def query_filter_default(cls, info):
                return True

            cls.query_filter = query_filter_default

        if folder:
            cls.folder = folder
        if get_perm:
            cls.get_permission = get_perm
        else:

            async def get_permission(module_class, context):
                return True

            cls.get_permission = get_permission

    @classmethod
    def get_auth(cls, module, request_type):
        class IsAuthenticated(BasePermission):
            async def has_permission(self, source, info, **kwargs) -> bool:
                if not await module.permission_getter(
                    info
                ) or not await cls.get_permission(
                    module, info.context, request_type
                ):
                    info.context['response'].status_code = 403
                    if not 'errors' in info.context['request'].scope:
                        info.context['request'].scope['errors'] = [module]
                    elif not module in info.context['request'].scope['errors']:
                        info.context['request'].scope['errors'].append(module)
                return True

        return IsAuthenticated



from datetime import date
from typing import Annotated, Optional

import strawberry
from sqlmodel import SQLModel, literal

from .setup import Setup


def set_dl(func):
    if not func.dl:
        return func
    f = func.dl + 'Filter'
    s = func.dl + 'Schema'
    if not func.many:

        async def new_func(
            self,
            info,
            filters: Annotated[f, strawberry.lazy('graphemy.router')]
            | None = None,
        ) -> Annotated[s, strawberry.lazy('graphemy.router')] | None:
            return await func(
                self,
                info,
                {
                    'filters': vars(filters) if filters else None,
                    'list_filters': None,
                },
            )

    else:

        async def new_func(
            self,
            info,
            filters: Annotated[f, strawberry.lazy('graphemy.router')]
            | None = None,
            list_filters: ListFilters | None = None,
        ) -> list[Annotated[s, strawberry.lazy('graphemy.router')]]:
            return await func(
                self,
                info,
                {
                    'filters': vars(filters) if filters else None,
                    'list_filters': vars(list_filters)
                    if list_filters
                    else None,
                },
            )

    new_func.__name__ = func.__name__
    return new_func


@strawberry.input
class ListFilters:
    field: str
    asc: bool


@strawberry.input
class MyDate:
    range: list[date] | None = None
    items: list[date] | None = None
    year: int | None = None


class Graphemy(SQLModel):
    _schema = None
    _query = None
    _filter = None
    __customfields__ = {}
    _default_mutation: bool = False
    _delete_mutation: bool = False
    _disable_query: bool = False

    async def permission_getter(info):
        return True

    @classmethod
    def auth(cls, request_type):
        return Setup.get_auth(cls, request_type)

    @classmethod
    @property
    def filter(cls):
        if cls._filter:
            return cls._filter.default
        return None

    @classmethod
    def query(cls):
        class Filter:
            pass

        for field_name, field in cls.__annotations__.items():
            field = (
                MyDate
                if (field == date or field == (date | None))
                else list[field]
            )
            setattr(
                Filter,
                field_name,
                strawberry.field(default=None, graphql_type=Optional[field]),
            )
        filter = strawberry.input(name=f'{cls.__name__}Filter')(Filter)
        cls._filter.default = filter

        async def field(
            self, info, filters: filter | None = None
        ) -> list[cls.schema]:
            if not await cls.permission_getter(
                info
            ) or not await Setup.get_permission(cls, info.context, 'query'):
                return []
            data = await get_all(cls, filters, Setup.query_filter(cls, info))
            return data

        return field

    @classmethod
    @property
    def schema(cls):
        return cls._schema

    @classmethod
    def set_schema(cls, classes):
        class Schema:
            pass

        for funcao in [
            func for func in cls.__dict__.values() if hasattr(func, 'dl')
        ]:
            setattr(
                Schema,
                funcao.__name__,
                strawberry.field(
                    set_dl(funcao),
                    permission_classes=[
                        Setup.get_auth(
                            classes[funcao.dl if funcao.dl else cls.__name__][
                                0
                            ],
                            'query',
                        )
                    ],
                ),
            )
        for field, v in cls.__customfields__.items():
            setattr(Schema, field, v)
        cls._schema = strawberry.experimental.pydantic.type(
            cls, all_fields=True, name=f'{cls.__name__}Schema'
        )(Schema)

    @classmethod
    @property
    def mutation(cls):
        from sqlalchemy.inspection import inspect as insp

        pk = [pk.name for pk in insp(cls).primary_key]

        class Filter:
            pass

        for field_name, field in cls.__annotations__.items():
            setattr(
                Filter,
                field_name,
                strawberry.field(default=None, graphql_type=Optional[field]),
            )
        input = strawberry.input(name=f'{cls.__name__}Input')(Filter)

        async def mutation(self, params: input) -> cls.schema:
            return await put_item(cls, params, pk)

        return mutation

    @classmethod
    @property
    def delete_mutation(cls):
        from sqlalchemy.inspection import inspect as insp

        pk = [pk.name for pk in insp(cls).primary_key]

        class Filter:
            pass

        for field_name, field in cls.__annotations__.items():
            if field_name in pk:
                setattr(
                    Filter,
                    field_name,
                    strawberry.field(
                        default=None, graphql_type=Optional[field]
                    ),
                )
        input = strawberry.input(name=f'{cls.__name__}InputPk')(Filter)

        async def mutation(self, params: input) -> cls.schema:
            return await delete_item(cls, params, pk)

        cls._delete = mutation
        return mutation


from datetime import date

from sqlmodel import Session, and_, bindparam, extract, or_, select

engine = Setup.engine


def get_keys(model: 'Graphemy', id: str | list[str]) -> tuple | str:
    """
    Retrieve one or multiple attributes or keys from a Graphemy instance.

    Args:
                    model (Graphemy): An instance of the Graphemy class from which attributes/keys will be retrieved.
                    id (str or list of str): The attribute/key name(s) to be retrieved from the model.

    Returns:
                    str, tuple, or any: The retrieved attribute(s) or key(s) from the model. If 'id' is a single string,the corresponding attribute/key value is returned. If 'id' is a list of strings, a tuple containing the corresponding attribute/key values in the order specified is returned. The returned values may be converted to strings if they are integers or have leading/trailing whitespaces.

    Examples:
                    >>> from graphemy import Graphemy

                    >>> class Hero(Graphemy):
                    ...     name:str
                    ...     power_level:int

                    >>> hero_instance = Hero(name='Superman', power_level=100)

                    >>> get_keys(hero_instance, 'name')
                    'Superman'
                    >>> get_keys(hero_instance, 'power_level')
                    100
                    >>> get_keys(hero_instance, ['name', 'power_level'])
                    ('Superman', 100)
    """
    if isinstance(id, list):
        return tuple([getattr(model, id[i]) for i in range(len(id))])
    value = getattr(model, id)
    if isinstance(value, str):
        value = value.strip()
    return value


def get_filter(
    model: 'Graphemy',
    keys,
    id,
    params: dict,
    i: int,
    contains: bool = False,
    contains_fore: bool = False,
):
    if isinstance(id, list):
        filter = []
        for j, key in enumerate(keys):
            f = []
            if None not in key:
                for k in range(len(id)):
                    f.append(
                        getattr(model, id[k])
                        == bindparam(
                            f'p{i}_{j}_{k}',
                            literal_execute=not isinstance(key[k], date),
                        )
                    )
                    params[f'p{i}_{j}_{k}'] = key[k]
                filter.append(and_(*f))
            else:
                filter.append(False)
        return or_(*filter)
    elif None in keys:
        keys.remove(None)
    params[f'p{i}'] = keys
    if contains:
        return or_(*[getattr(model, id).contains(valor) for valor in keys])
    if contains_fore:
        return or_(
            *[
                literal(value).like('%' + getattr(model, id) + '%')
                for value in keys
            ]
        )
    return getattr(model, id).in_(
        bindparam(f'p{i}', expanding=True, literal_execute=True)
    )


async def get_list(
    model: 'Graphemy',
    parameters: list[tuple],
    id='id',
    contains=False,
    contains_fore=False,
):

    groups = {}
    id_groups = {}
    filters = {}
    params = {}
    for p in parameters:
        f = tuple([p[0], p[2]])
        if not f in filters:
            filters[f] = []
        filters[f].append(p[1][1])
        groups[p] = []
        if not p[1][1] in id_groups:
            id_groups[p[1][1]] = []
        id_groups[p[1][1]].append(p)
    query = select(model)
    query_filters = []
    for i, f in enumerate(filters):
        if f[0][1]:
            filter_temp = []
            for k, v in f[0][1]:
                if v:
                    if isinstance(v[0], tuple):
                        if v[2][1]:
                            filter_temp.append(
                                extract('year', getattr(model, k)) == v[2][1]
                            )
                        if v[0][1]:
                            filter_temp.append(
                                getattr(model, k).in_(list(v[0][1]))
                            )
                        if v[1][1] and v[1][1][0]:
                            filter_temp.append(getattr(model, k) >= v[1][1][0])
                        if v[1][1] and v[1][1][1]:
                            filter_temp.append(getattr(model, k) <= v[1][1][1])
                    else:
                        filter_temp.append(getattr(model, k).in_(v))
        else:
            filter_temp = [True]

        query_filters.append(
            and_(
                *filter_temp,
                get_filter(
                    model,
                    filters[f],
                    id,
                    params,
                    i,
                    contains=contains,
                    contains_fore=contains_fore,
                ),
            )
        )
    results = await Setup.execute_query(
        query.where(or_(*query_filters)).params(**params)
    )
    for r in results:
        if contains:
            key = get_keys(r, id)
            for i in id_groups:
                if i in key:
                    for t in id_groups[i]:
                        if not t[0][1] or all(
                            [getattr(r, k) in v for k, v in t[0][1] if v]
                        ):
                            groups[t].append(r)
        elif contains_fore:
            temp = get_keys(r, id)
            for key, v in id_groups.items():
                if temp in key:
                    for t in v:
                        groups[t].append(r)

        else:
            temp = id_groups[get_keys(r, id)]
            if len(temp) == 1:
                key = temp[0]
                groups[key].append(r)
            else:
                for t in temp:
                    if not t[0][1] or all(
                        [getattr(r, k) in v for k, v in t[0][1] if v]
                    ):
                        groups[t].append(r)
    return groups.values()


async def get_one(model: 'Graphemy', parameters: list[tuple], id='id'):
    groups = {}
    id_groups = {}
    filters = {}
    params = {}
    for p in parameters:
        f = tuple([p[0], p[2]])
        if not f in filters:
            filters[f] = []
        filters[f].append(p[1][1])
        groups[p] = None
        if not p[1][1] in id_groups:
            id_groups[p[1][1]] = []

        id_groups[p[1][1]].append(p)
    query = select(model)
    query_filters = []
    for i, f in enumerate(filters):
        if f[0][1]:
            filter_temp = []
            for k, v in f[0][1]:
                if v:
                    if isinstance(v[0], tuple):
                        if v[2][1]:
                            filter_temp.append(
                                extract('year', getattr(model, k)) == v[2][1]
                            )
                        if v[0][1]:
                            filter_temp.append(
                                getattr(model, k).in_(list(v[0][1]))
                            )
                        if v[1][1] and v[1][1][0]:
                            filter_temp.append(getattr(model, k) >= v[1][1][0])
                        if v[1][1] and v[1][1][1]:
                            filter_temp.append(getattr(model, k) <= v[1][1][1])
                    else:
                        filter_temp.append(getattr(model, k).in_(v))
        else:
            filter_temp = [True]

        query_filters.append(
            and_(*filter_temp, get_filter(model, filters[f], id, params, i))
        )
    results = await Setup.execute_query(
        query.where(or_(*query_filters)).params(**params)
    )
    for r in results:
        temp = id_groups[get_keys(r, id)]
        if len(temp) == 1:
            key = temp[0]
            groups[key] = r
        else:
            for t in temp:
                if not t[0][1] or all(
                    [getattr(r, k) in v for k, v in t[0][1] if v]
                ):
                    groups[t] = r
    return groups.values()


async def get_all(model: 'Graphemy', filters, query_filter):
    query = select(model).where(query_filter)
    if filters:
        filters = vars(filters)
        for k, v in filters.items():
            if isinstance(v, MyDate):
                if v.year:
                    query = query.where(
                        extract('year', getattr(model, k)) == v.year
                    )
                if v.items:
                    query = query.where(getattr(model, k).in_(v.items))
                if v.range and v.range[0]:
                    query = query.where(getattr(model, k) >= (v.range[0]))
                if v.range and v.range[1]:
                    query = query.where(getattr(model, k) <= (v.range[1]))
            elif filters[k]:
                query = query.where(getattr(model, k).in_(filters[k]))
    r = await Setup.execute_query(query)
    return r


async def put_item(model: 'Graphemy', item, id='id', engine=engine):
    id = [getattr(item, i) for i in id]
    kwargs = vars(item)
    engine = Setup.engine
    with Session(engine) as session:
        if not id or None in id:
            new_item = model(**kwargs)
        else:
            new_item = session.get(model, id)
            if not new_item:
                new_item = model(**kwargs)
            for key, value in kwargs.items():
                setattr(new_item, key, value)
        session.add(new_item)
        session.commit()
        session.refresh(new_item)
    return new_item


async def delete_item(model: 'Graphemy', item, id='id', engine=engine):
    id = [getattr(item, i) for i in id]
    engine = Setup.engine
    with Session(engine) as session:
        item = session.get(model, id)
        session.delete(item)
        session.commit()
    return item
